import logging
import re
from dataclasses import dataclass, field

from rich.console import Console

__all__ = (
    "logger",
    "init_logging",
    "TAG_START",
    "TAG_END",
    "BRANCH_START",
    "BRANCH_END",
    "COMMIT_START",
    "COMMIT_END",
    "DRY_RUN_START",
    "DRY_RUN_END",
    "RELEASE_TAG_START",
    "RELEASE_TAG_END",
)

_logger = logging.getLogger("track-bump")


def init_logging(level: int = logging.WARNING):
    stream_handler = logging.StreamHandler()
    formatter = logging.Formatter("%(asctime)s [%(levelname)s] %(message)s", datefmt="%Y-%m-%d %H:%M:%S")
    stream_handler.setFormatter(formatter)
    _logger.addHandler(stream_handler)
    _logger.setLevel(level)


@dataclass
class RichLogger:
    logger: logging.Logger
    console: Console = field(default_factory=Console)

    @property
    def level(self):
        return self.logger.level

    @level.setter
    def level(self, value):
        self.logger.setLevel(value)

    @property
    def disabled(self) -> bool:
        return self.logger.disabled

    @disabled.setter
    def disabled(self, value: bool):
        self.logger.disabled = value

    def can_print(self, level: int) -> bool:
        return self.level <= level and self.disabled

    def _log(
        self,
        level: int,
        msg: str,
        *args,
        **kwargs,
    ):
        if not self.disabled:
            msg = rm_markdown(msg)
        self.logger.log(level, msg, *args, **kwargs)

    def debug(self, msg: str, *args, **kwargs):
        self._log(logging.DEBUG, msg, *args, **kwargs)
        if self.can_print(logging.DEBUG):
            self.console.print(f"[steel_blue]{msg}[/steel_blue]")

    def info(self, msg: str, *args, **kwargs):
        self._log(logging.INFO, msg, *args, **kwargs)
        if self.can_print(logging.INFO):
            self.console.print(msg)

    def warning(self, msg: str, *args, **kwargs):
        self._log(logging.WARNING, msg, *args, **kwargs)
        if self.can_print(logging.WARNING):
            self.console.print(f"[yellow]{msg}[/yellow]")

    def error(self, msg: str, *args, **kwargs):
        self._log(logging.ERROR, msg, *args, **kwargs)
        if self.can_print(logging.ERROR):
            self.console.print(f"[red]{msg}[/red]")


logger = RichLogger(logger=_logger)

_MARKDOWN_PATTERN = re.compile(r"\[\/?\w+(?:=[^\]]*)?\]")


def rm_markdown(text: str) -> str:
    """
    Removes rich-style or markdown-like tags from the input string.
    It removes tags such as [bold], [italic], [color], etc.
    """
    return _MARKDOWN_PATTERN.sub("", text)


TAG_START = "[bold][cyan]"
TAG_END = "[/cyan][/bold]"

BRANCH_START = "[bold]"
BRANCH_END = "[/bold]"

COMMIT_START = "[bold][blue]"
COMMIT_END = "[/blue][/bold]"

DRY_RUN_START = "[grey58]"
DRY_RUN_END = "[/grey58]"

RELEASE_TAG_START = "[bold][green]"
RELEASE_TAG_END = "[/green][/bold]"



import logging
import os
from pathlib import Path

import pytest

STATIC_DIR = Path(__file__).parent / "static"
DEFAULT_BRANCH = "master"
os.environ["DEFAULT_BRANCH"] = DEFAULT_BRANCH
os.environ["CI_USER"] = "foo"
os.environ["CI_USER_EMAIL"] = "foo@bar.com"


@pytest.fixture(scope="session", autouse=True)
def setup_logging():
    from track_bump.logs import init_logging, logger

    init_logging(logging.WARNING)
    logger.console.quiet = True



import textwrap
import tomllib
from contextlib import nullcontext

import pytest


@pytest.mark.parametrize(
    "filename, config, expected",
    [
        # TOML
        pytest.param(
            "pyproject.toml",
            """
            [tool.track-bump]
            version = "0.1.0"
            bump_message = "foo"
            """,
            nullcontext({"version": "0.1.0", "bump_message": "foo", "version_files": []}),
            id="toml - valid config",
        ),
        pytest.param(
            "pyproject.toml",
            """
            foo
            """,
            pytest.raises(tomllib.TOMLDecodeError),
            id="toml - invalid file",
        ),
        pytest.param(
            "pyproject.toml",
            """
            [tool.track-bump]
            version = "0.1.0"
            """,
            pytest.raises(ValueError, match="bump_message is required in config file"),
            id="toml - missing bump_message",
        ),
        pytest.param(
            "pyproject.toml",
            """
            [tool.track-bump]
            bump_message = "foo"
            """,
            pytest.raises(ValueError, match="version is required in config file"),
            id="toml - missing version",
        ),
        pytest.param(
            "pyproject.toml",
            """
            [tool.track-bump]
            version = "0.1.0"
            bump_message = "foo"
            version_files = ["pyproject.toml"]
            """,
            nullcontext({"version": "0.1.0", "bump_message": "foo", "version_files": ["pyproject.toml"]}),
            id="toml - valid config with version_files",
        ),
        # JSON
        pytest.param(
            "package.json",
            """
            {
                "version": "0.1.0",
                "track-bump": {
                    "bumpMessage": "foo"
                }
            }
            """,
            nullcontext({"version": "0.1.0", "bump_message": "foo", "version_files": []}),
            id="json - valid config",
        ),
        pytest.param(
            "package.json",
            """
            {
                "version": "0.1.0",
                "track-bump": {
                }
            }
            """,
            pytest.raises(ValueError, match="bump_message is required in config file"),
            id="json - missing bump_message",
        ),
    ],
)
def test_load_config(tmp_path, filename, config, expected):
    from track_bump.config import Config

    _path = tmp_path / filename
    _path.write_text(config)
    with expected as e:
        _config = Config.from_file(_path)
        assert _config.version == e["version"]
        assert _config.bump_message == e["bump_message"]
        assert _config.version_files == e["version_files"]


@pytest.mark.parametrize(
    "filename, file_content,params,expected",
    [
        pytest.param(
            "pyproject.toml",
            """
            [tool.track-bump]
            version = "0.1.0"
            bump_message = "foo"
            """,
            {"version": "0.2.0", "tag": "version"},
            """
            [tool.track-bump]
            version = "0.2.0"
            bump_message = "foo"
            """,
            id="toml - valid",
        ),
        pytest.param(
            "package.json",
            """
            {
                "version": "0.1.0",
                "track-bump": {
                    "bumpMessage": "foo"
                }
            }
            """,
            {"version": "0.2.0", "tag": "version"},
            """
            {
                "version": "0.2.0",
                "track-bump": {
                    "bumpMessage": "foo"
                }
            }
            """,
            id="json - valid",
        ),
    ],
)
def test_replace_in_file(tmp_path, filename, expected, file_content, params):
    from track_bump.config import replace_in_file

    _path = tmp_path / filename
    _path.write_text(textwrap.dedent(file_content))

    replace_in_file(_path, **params)

    assert _path.read_text() == textwrap.dedent(expected)



import pytest


@pytest.mark.parametrize(
    "version, expected",
    [
        pytest.param("0.1.0", ((0, 1, 0), None), id="0.1.0"),
        pytest.param("v0.1.0", ((0, 1, 0), None), id="v0.1.0"),
        pytest.param("0.1.0-beta.1", ((0, 1, 0), ("beta", 1)), id="0.1.0-beta.1"),
    ],
)
def test_parse_version(version, expected):
    from track_bump.utils import parse_version

    assert parse_version(version) == expected



import re

from .logs import COMMIT_END, COMMIT_START, logger
from .utils import get_last_tag, parse_version

__all__ = (
    "get_latest_stable_tag",
    "get_latest_release_tag",
    "get_branch_release",
    "get_new_tag",
)


def get_latest_stable_tag():
    f"""
    Get the latest tag of the DEFAULT_BRANCH branch (stable)
    For example:
     - if the DEFAULT_BRANCH has a tag v0.1.0, it will return v0.1.0
    """
    return get_last_tag(r"^v\d+\.\d+\.\d+$")


def get_latest_release_tag(release_tag: str) -> str | None:
    """
    Get the latest tag of the given release_tag
    For example:
        - if the release_tag is "beta", it will return the latest tag v0.1.0-beta.1
    """
    return get_last_tag(rf"^v\d+\.\d+\.\d+-{release_tag}\.\d+$")


def get_branch_release(branch: str, releases: dict[str, str]) -> str:
    """
    Get the release name for the given branch
    """

    for branch_pattern, release_tag in releases.items():
        if re.match(branch_pattern, branch):
            return release_tag

    _supported_branches = ", ".join(releases.keys())
    raise ValueError(f"Branch {branch!r} is not supported. Supported branches are: {_supported_branches}")


_BUMP_MINOR_REG = re.compile(r"release:.*")


def get_new_tag(
    stable_tag: str,
    release_tag: str | None,
    release: str,
    last_commit_message: str | None = None,
) -> str:
    """
    Return the new tag based on the latest release tag and current branch
    Rules are:
     - if branch is not stable:
        - if the latest stable tag is v0.1.0 and the latest release tag is v0.2.0-beta.1:  v0.2.0-beta.2
        - if the latest stable tag is v0.1.0 and there is no latest release tag: v0.2.0-beta.0
    - if branch is stable and the latest stable tag is v0.1.0:
        - if the last commit message is "release: .*": v0.2.0
        - else: v0.1.1
    """
    (major, minor, patch), _ = parse_version(stable_tag)
    _next_release = f"v{major}.{minor + 1}.0"
    # We are releasing a new version
    if release == "stable":
        logger.info(
            f"Last commit message: {COMMIT_START}{last_commit_message}{COMMIT_END}"
            if last_commit_message is not None
            else "No commit message found"
        )
        if last_commit_message is None or _BUMP_MINOR_REG.match(last_commit_message):
            logger.debug("Bumping minor")
            _tag = _next_release
        else:
            logger.debug("Bumping patch")
            _tag = f"v{major}.{minor}.{patch + 1}"
    else:
        if release_tag is not None:
            (_release_major, _release_minor, _release_patch), _tag_part = parse_version(release_tag)
            if _tag_part is None:
                raise ValueError(f"Invalid tag: {release_tag!r}")
            (_tag, _tag_version) = _tag_part
            if f"{major}.{minor}.{patch}" == f"{_release_major}.{_release_minor}.{_release_patch}":
                _release_number = 0
            else:
                _release_number = _tag_version + 1
        else:
            _release_number = 0
        _tag = f"{_next_release}-{release}.{_release_number}"

    return _tag



from contextlib import nullcontext

import pytest

from track_bump.config import get_default_releases

from .conftest import DEFAULT_BRANCH


@pytest.mark.parametrize(
    "branch, expected",
    [
        pytest.param("develop", nullcontext("beta"), id="develop"),
        pytest.param(DEFAULT_BRANCH, nullcontext("stable"), id="main branch"),
        pytest.param("release/foo", nullcontext("rc"), id="release"),
        pytest.param("foo", pytest.raises(ValueError, match="Branch 'foo' is not supported"), id="invalid branch"),
    ],
)
def test_get_branch_release(branch, expected):
    from track_bump.tags import get_branch_release

    with expected as e:
        assert get_branch_release(branch, releases=get_default_releases()) == e


@pytest.mark.parametrize(
    "params, expected",
    [
        pytest.param(
            {"stable_tag": "v0.1.0", "release_tag": None, "release": "beta"}, nullcontext("v0.2.0-beta.0"), id="beta"
        ),
        pytest.param(
            {"stable_tag": "v0.1.0", "release_tag": "v0.2.0-beta.1", "release": "beta"},
            nullcontext("v0.2.0-beta.2"),
            id="beta existing release tag",
        ),
        pytest.param(
            {"stable_tag": "v0.1.0", "release_tag": None, "release": "stable"},
            nullcontext("v0.2.0"),
            id="stable",
        ),
        pytest.param(
            {
                "stable_tag": "v0.1.0",
                "release_tag": None,
                "release": "stable",
                "last_commit_message": "release: v0.1.0",
            },
            nullcontext("v0.2.0"),
            id="stable with commit message",
        ),
        pytest.param(
            {"stable_tag": "v0.1.0", "release_tag": None, "release": "stable", "last_commit_message": "fix: v0.1.0"},
            nullcontext("v0.1.1"),
            id="stable patch with commit message",
        ),
    ],
)
def test_get_new_tag(params, expected):
    from track_bump.tags import get_new_tag

    with expected as e:
        assert get_new_tag(**params) == e



import json
import re
import tomllib
from dataclasses import dataclass, field
from pathlib import Path

from .logs import logger

__all__ = ("Config", "replace_in_file", "replace_in_files")

from track_bump import env


def get_default_releases() -> dict[str, str]:
    return {
        r"^develop$": "beta",
        r"^release/.*": "rc",
        rf"^{env.DEFAULT_BRANCH}$": "stable",
    }


def replace_in_file(file_path: Path, version: str, tag: str):
    """
    Replace the tag with the new version in the given file
    For example, if the file contains:
        version = "0.1.0"
    and you call replace_in_file(file_path, "0.2.0", "version")
    The file will be updated to:
        version = "0.2.0"
    """
    match file_path.suffix:
        case ".toml":
            _new_text = re.sub(rf'^{tag} = "(.*)"$', rf'{tag} = "{version}"', file_path.read_text(), flags=re.MULTILINE)
        case ".json":
            _new_text = re.sub(rf'"({tag})": "(.*)"', rf'"{tag}": "{version}"', file_path.read_text())
        case _:
            raise ValueError(f"Only .toml and .json files are supported")
    file_path.write_text(_new_text)


CONFIG_FILES = [".cz.toml", "pyproject.toml", "package.json"]


@dataclass
class Config:
    version: str
    bump_message: str
    version_files: list[str]
    default_branch: str
    releases: dict[str, str] = field(default_factory=get_default_releases)

    _config_path: Path = field(init=False)

    # def __post_init__(self):
    #     if self.default_branch not in self.releases:
    #         raise ValueError(f"Default branch {self.default_branch!r} is not supported in releases."
    #                          f" Supported branches are: {', '.join(self.releases.keys())}")

    @property
    def config_path(self) -> Path:
        return self._config_path

    @property
    def project_path(self) -> Path:
        return self._config_path.parent

    @classmethod
    def from_file(cls, config_path: Path, default_branch: str = env.DEFAULT_BRANCH):
        if not config_path.exists():
            raise FileNotFoundError(f"{config_path} not found")
        logger.debug(f"Parsing {config_path}")

        if config_path.suffix == ".toml":
            data = tomllib.loads(config_path.read_text())
            _config = data.get("tool", {}).get("track-bump")
            if _config is None:
                raise ValueError(f"Could not find config tool.track-bump in {config_path!r}")
            version = _config.get("version")
        elif config_path.suffix == ".json":
            data = json.loads(config_path.read_text())
            _config = data.get("track-bump")
            if _config is None:
                raise ValueError(f"Could not find config track-bump in {config_path}")
            version = data.get("version")
        else:
            raise ValueError("Only .toml and .json files are supported")

        if version is None:
            raise ValueError("version is required in config file")
        bump_message = _config.get("bump_message") or _config.get("bumpMessage")
        if bump_message is None:
            raise ValueError("bump_message is required in config file")

        version_files = _config.get("version_files") or _config.get("versionFiles") or []

        releases = _config.get("releases") or get_default_releases()
        config = cls(
            version=version,
            bump_message=bump_message,
            version_files=version_files,
            releases=releases,
            default_branch=default_branch,
        )
        config._config_path = config_path
        return config

    @classmethod
    def from_project(cls, project_path: Path, default_branch: str = env.DEFAULT_BRANCH):
        # Check if any of the config files exist
        for file in CONFIG_FILES:
            config_path = Path(project_path / file)
            if config_path.exists():
                logger.debug(f"Found config file: {config_path}")
                break
        else:
            raise FileNotFoundError(f"Could not find any of the following files: {CONFIG_FILES} in {project_path}")

        return cls.from_file(config_path, default_branch=default_branch)


def replace_in_files(config_path: Path, files: list[str], version: str):
    """
    Replace the version in the given files
    """
    for _file in files:
        try:
            _path, _tag = _file.split(":")
        except ValueError:
            _path = _file
            _tag = "version"
        _file_path = Path(config_path.parent / _path)
        if not _file_path.exists():
            raise FileNotFoundError(f"{_file_path} not found")
        logger.debug(f"Replacing in {_file_path}")
        replace_in_file(_file_path, version=version, tag=_tag)



import logging
from pathlib import Path

from piou import Cli, Option

from .bump import bump_project
from .config import Config
from .logs import init_logging as _init_logging
from .logs import logger
from .tags import get_branch_release, get_latest_release_tag, get_latest_stable_tag
from .utils import get_current_branch, set_cd

cli = Cli("Track-bump utility commands")

cli.add_option("-v", "--verbose", help="Verbosity")
cli.add_option("-vv", "--verbose2", help="Increased verbosity")
cli.add_option("--init-logging", help="Initialize logging")


def on_process(verbose: bool = False, verbose2: bool = False, init_logging: bool = False):
    logger.level = logging.DEBUG if verbose2 else logging.INFO if verbose else logging.WARNING
    logger.disabled = not init_logging
    if init_logging:
        _init_logging(logger.level)


cli.set_options_processor(on_process)


@cli.command(cmd="bump", help="Bump project version")
def bump(
    project_path: Path = Option(Path.cwd(), "-p", "--project", help="Project path"),
    sign_commits: bool = Option(False, "--sign", help="Sign commits"),
    branch: str | None = Option(None, "--branch", help="Branch to bump"),
    dry_run: bool = Option(False, "--dry-run", help="Dry run"),
    force: bool = Option(False, "--force", help="Force fetch tags"),
    no_reset_git: bool = Option(False, "--no-reset-git", help="Do not reset git config"),
    no_tag: bool = Option(False, "--no-tag", help="Do not create a tag"),
    pre_release: str | None = Option(None, "--pre-release", help="Pre-release version"),
):
    """
    Bump the version of the project:
    - Find the latest tag of the given branch
    - Bump the version based on the branch
       - Creates the tag
       - Update the version files
    - Commit the changes and tag

    The branches are mapped to the release tags as follows:
    - develop: beta
    - release: rc
    """
    config = Config.from_project(project_path)
    bump_project(
        config,
        sign_commits,
        branch=branch,
        dry_run=dry_run,
        force=force,
        no_reset_git=no_reset_git,
        add_tag=not no_tag,
        pre_release=pre_release,
    )


@cli.command(cmd="get-latest-tag", help="Get the latest tag")
def get_latest_tag(
    project_path: Path = Option(Path.cwd(), "-p", "--project", help="Project path"),
    branch: str | None = Option(None, "--branch", help="Branch to bump"),
    pre_release: str | None = Option(None, "--pre-release", help="Pre-release version"),
):
    f"""
    Prints the latest tag for the given branch (default: current branch)
    otherwise, it will return the latest associated release tag
    """
    config = Config.from_project(project_path)
    with set_cd(project_path):
        _branch = branch or get_current_branch()
        logger.info(f"Getting latest tag for branch {_branch}")
        _release = pre_release or get_branch_release(_branch, releases=config.releases)
        tag = get_latest_stable_tag() if _branch == config.default_branch else get_latest_release_tag(_release)
    if tag:
        print(tag)


def run():
    cli.run()


if __name__ == "__main__":
    run()



import json
import shutil
import tomllib
from pathlib import Path

import pytest

from .conftest import DEFAULT_BRANCH, STATIC_DIR


def get_tags(project_path: Path):
    from track_bump.utils import get_tags, set_cd

    with set_cd(project_path):
        tags = get_tags()
    return tags


def get_toml_versions(project_path: Path):
    default_version = tomllib.loads((project_path / ".cz.toml").read_text())["tool"]["track-bump"]["version"]
    _sub_1 = tomllib.loads((project_path / "sub-project-1" / "pyproject.toml").read_text())["tool"]["poetry"]["version"]
    _sub_2 = tomllib.loads((project_path / "sub-project-2" / "pyproject.toml").read_text())["tool"]["poetry"]["version"]
    return {"default": default_version, "sub_1": _sub_1, "sub_2": _sub_2}


def get_js_versions(project_path: Path):
    default_version = json.loads((project_path / "package.json").read_text())["version"]
    return {"default": default_version}


class TestTOML:
    @pytest.fixture(scope="function")
    def project_path(self, tmp_path: Path):
        return tmp_path / "project"

    @pytest.fixture(scope="function")
    def setup_project(self, project_path: Path):
        shutil.copytree(STATIC_DIR / "project", project_path)
        from track_bump.utils import exec_cmd, git_setup, set_cd

        with set_cd(project_path):
            exec_cmd("git init")
            with git_setup(sign_commits=False, default_branch=DEFAULT_BRANCH):
                yield

    def test_bump(self, setup_project, project_path, monkeypatch):
        monkeypatch.setattr("track_bump.utils.get_last_commit_message", lambda: None)
        from track_bump.bump import bump_project
        from track_bump.config import Config

        config = Config.from_project(project_path)

        assert get_toml_versions(project_path) == {
            "default": "0.1.0",
            "sub_1": "0.1.0",
            "sub_2": "0.1.0",
        }, "Initial versions should be 0.1.0"
        assert get_tags(project_path) == [], "No tags should be present"

        bump_project(config=config, branch="develop")
        tags = ["v0.2.0-beta.0"]
        assert set(get_tags(project_path)) == set(tags)
        assert get_toml_versions(project_path) == {
            "default": "0.2.0-beta.0",
            "sub_1": "0.2.0-beta.0",
            "sub_2": "0.2.0-beta.0",
        }
        # Running it again
        tags.append("v0.2.0-beta.1")
        bump_project(config=config, branch="develop")
        assert set(get_tags(project_path)) == set(tags), "Tags should be incremented"
        assert get_toml_versions(project_path) == {
            "default": "0.2.0-beta.1",
            "sub_1": "0.2.0-beta.1",
            "sub_2": "0.2.0-beta.1",
        }

        # Changing to main, making a fix

        bump_project(config=config, last_commit_message="fix: a fix", branch=DEFAULT_BRANCH)
        tags.append("v0.1.1")
        assert set(get_tags(project_path)) == set(tags)
        assert get_toml_versions(project_path) == {"default": "0.1.1", "sub_1": "0.1.1", "sub_2": "0.1.1"}

        # Creating a release

        bump_project(config=config, branch=DEFAULT_BRANCH)
        tags.append("v0.2.0")
        assert set(get_tags(project_path)) == set(tags)
        assert get_toml_versions(project_path) == {"default": "0.2.0", "sub_1": "0.2.0", "sub_2": "0.2.0"}


class TestJSON:
    @pytest.fixture(scope="function")
    def project_path(self, tmp_path: Path):
        return tmp_path / "project-js"

    @pytest.fixture(scope="function")
    def setup_project(self, project_path: Path):
        shutil.copytree(STATIC_DIR / "project-js", project_path)
        from track_bump.utils import exec_cmd, git_setup, set_cd

        with set_cd(project_path):
            exec_cmd("git init")
            with git_setup(sign_commits=False, default_branch=DEFAULT_BRANCH):
                yield

    def test_bump(self, setup_project, project_path, monkeypatch):
        monkeypatch.setattr("track_bump.utils.get_last_commit_message", lambda: None)
        from track_bump.bump import bump_project
        from track_bump.config import Config

        config = Config.from_project(project_path)

        assert get_js_versions(project_path) == {
            "default": "0.1.0",
        }, "Initial versions should be 0.1.0"
        assert get_tags(project_path) == [], "No tags should be present"

        bump_project(config=config, branch="develop")
        tags = ["v0.2.0-beta.0"]
        assert set(get_tags(project_path)) == set(tags)
        assert get_js_versions(project_path) == {"default": "0.2.0-beta.0"}
        # Running it again
        tags.append("v0.2.0-beta.1")
        bump_project(config=config, branch="develop")
        assert set(get_tags(project_path)) == set(tags), "Tags should be incremented"
        assert get_js_versions(project_path) == {"default": "0.2.0-beta.1"}

        # Changing to main, making a fix

        bump_project(config=config, last_commit_message="fix: a fix")
        tags.append("v0.1.1")
        assert set(get_tags(project_path)) == set(tags)
        assert get_js_versions(project_path) == {"default": "0.1.1"}

        # Creating a release

        bump_project(config=config)
        tags.append("v0.2.0")
        assert set(get_tags(project_path)) == set(tags)
        assert get_js_versions(project_path) == {"default": "0.2.0"}



from pathlib import Path

from track_bump.config import Config, replace_in_files
from track_bump.tags import get_branch_release, get_latest_release_tag, get_latest_stable_tag, get_new_tag
from track_bump.utils import (
    fetch_tags,
    get_current_branch,
    get_last_commit_message,
    git_commit,
    git_setup,
    git_tag,
    parse_version,
    set_cd,
)

from .logs import (
    COMMIT_END,
    COMMIT_START,
    DRY_RUN_END,
    DRY_RUN_START,
    TAG_END,
    TAG_START,
    logger,
)


def bump_project(
    config: Config,
    sign_commits: bool = False,
    branch: str | None = None,
    last_commit_message: str | None = None,
    dry_run: bool = False,
    force: bool = False,
    no_reset_git: bool = False,
    add_tag: bool = True,
    pre_release: str | None = None,
):
    """
    Bump the version of the project, create a commit and tag and commit the changes.
    You can also add files to be added to the commit.
    If add_tag is specified, it will also create a tag with the new version. Otherwise
    it'll just print the new tag.
    """
    # Setup git
    current_version = config.version
    with set_cd(config.project_path):
        with git_setup(sign_commits=sign_commits, no_reset=no_reset_git):
            # Get the latest stable and release tags for the branch
            fetch_tags(force=force)
            _latest_stable_tag = get_latest_stable_tag()
            _branch = branch or get_current_branch()
            _release = pre_release or get_branch_release(_branch, releases=config.releases)
            # If no latest tag, use the current version
            if _latest_stable_tag is None:
                (major, minor, path), release = parse_version(current_version)
                _latest_stable_tag = f"v{major}.{max(minor - 1, 1)}.{path}"

            _latest_release_tag = get_latest_release_tag(_release)
            _new_tag = get_new_tag(
                stable_tag=_latest_stable_tag,
                release_tag=_latest_release_tag,
                last_commit_message=last_commit_message or get_last_commit_message(),
                release=_release,
            )

            new_version = _new_tag.removeprefix("v")
            logger.info(
                f"Stable tag: {TAG_START}{_latest_stable_tag}{TAG_END} | "
                f"Latest release tag: {TAG_START}{_latest_release_tag}{TAG_END} | "
                f"New version: {new_version} "
                f"(branch: {_branch}, release: {_release})"
            )

            version_files = config.version_files + [f"{config.config_path.name}:version"]
            if not dry_run:
                replace_in_files(config.config_path, version_files, new_version)
            else:
                logger.info(
                    f"{DRY_RUN_START}Would replace version with {new_version} in files:\n - {'\n - '.join(version_files)}"
                )
            _bump_message = config.bump_message.format(current_version=current_version, new_version=new_version)
            if not dry_run:
                logger.info(f"Committing with message: {COMMIT_START}{_bump_message}{COMMIT_END}")
                git_commit(_bump_message)
                if add_tag:
                    git_tag(_new_tag)
            else:
                logger.info(
                    f"{DRY_RUN_START}Would commit with message: {COMMIT_START}{_bump_message}{COMMIT_END} "
                    f"and tag: {TAG_START}{_new_tag}{TAG_END}{DRY_RUN_END}"
                )
            logger.info("Done")
            if not add_tag:
                print(_new_tag)



import contextlib
import os
import pathlib
import re
import subprocess

from track_bump.env import CI_USER, CI_USER_EMAIL

from .logs import logger

__all__ = (
    "exec_cmd",
    "get_last_tag",
    "git_tag",
    "git_setup",
    "set_cd",
    "get_current_branch",
    "git_commit",
    "parse_version",
    "get_tags",
    "get_last_commit_message",
    "fetch_tags",
    "get_default_branch",
)


def exec_cmd(
    cmd: str | list[str], *, env: dict | None = None, show_progress: bool = False, ignore_errors: bool = False
) -> str:
    default_shell = os.getenv("SHELL", "/bin/bash")
    logger.debug(f"Executing command {cmd!r}")
    process = subprocess.Popen(
        cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, executable=default_shell, env=env, text=True
    )
    if show_progress:
        for line in process.stderr or []:
            logger.debug(f" {line.rstrip()}")

    stdout, stderr = process.communicate()
    exit_code = process.wait()
    if not ignore_errors and exit_code != 0:
        raise OSError(stderr)

    if stdout:
        logger.debug(f"Command output: {stdout!r}")
    return stdout


@contextlib.contextmanager
def set_cd(path: pathlib.Path):
    prev_cwd = pathlib.Path.cwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(prev_cwd)


def fetch_tags(force: bool = False):
    logger.debug(f"Fetching tags (force: {force})")
    exec_cmd("git fetch --tags" + (" --force" if force else ""))


def get_tags():
    tags = exec_cmd("git tag --sort=-version:refname").split("\n")
    return [x.strip() for x in tags if x.strip()]


def get_last_tag(pattern: str) -> str | None:
    _tags = get_tags()
    _valid_tags = [_tag for _tag in _tags if re.match(pattern, _tag)]
    return _valid_tags[0] if _valid_tags else None


def git_tag(version: str):
    _output = exec_cmd(f"git tag {version}")


@contextlib.contextmanager
def git_setup(sign_commits: bool = False, default_branch: str | None = None, no_reset: bool = False):
    _cached = {
        "user.email": get_git_email(ignore_errors=True),
        "user.name": get_git_user_name(ignore_errors=True),
        "init.defaultBranch": get_default_branch(ignore_errors=True),
    }

    _ci_user = CI_USER or get_git_user_name()
    if not _ci_user:
        raise ValueError("CI_USER must be set")

    _ci_email = CI_USER_EMAIL or get_git_email()
    if not _ci_email:
        raise ValueError("CI_USER_EMAIL must be set")

    exec_cmd(f'git config user.email "{CI_USER_EMAIL}"')
    exec_cmd(f'git config user.name "{CI_USER}"')
    if sign_commits:
        exec_cmd("git config commit.gpgSign true")
    if default_branch:
        exec_cmd(f'git config init.defaultBranch "{default_branch}"')
    yield
    if no_reset:
        return

    for key, value in _cached.items():
        if value:
            exec_cmd(f'git config {key} "{value}"')
        else:
            try:
                exec_cmd(f"git config --unset {key}")
            except OSError as e:
                logger.warning(f"Failed to run 'git config --unset {key}' ({e.args})")


def get_current_branch() -> str:
    return exec_cmd("git branch --show-current").strip()


def git_commit(message: str):
    exec_cmd("git add .")
    exec_cmd(f'git commit -m "{message}"')


def get_last_commit_message() -> str | None:
    _latest_commit = exec_cmd("git log -1 --pretty=%B").strip()
    return _latest_commit if _latest_commit else None


type MajorMinorPatch = tuple[int, int, int]
type ReleaseVersion = tuple[str, int]


def parse_version(version: str) -> tuple[MajorMinorPatch, ReleaseVersion | None]:
    """
    Parse the version string and return a tuple with the major, minor, patch and the release version if any
    For example:
    - v0.1.0-beta.1 -> ((0, 1, 0), ('beta', 1))
    - v0.1.0 -> ((0, 1, 0), None)
    """

    _version, *_release = version.removeprefix("v").split("-")
    major, minor, patch = [int(x) for x in _version.split(".")]

    if _release:
        _release_name, _release_number_str = _release[0].split(".")
        release = (_release_name, int(_release_number_str))
    else:
        release = None
    return (major, minor, patch), release


def get_git_email(ignore_errors: bool = False):
    return exec_cmd("git config user.email", ignore_errors=ignore_errors).strip()


def get_git_user_name(ignore_errors: bool = False):
    return exec_cmd("git config user.name", ignore_errors=ignore_errors).strip()


def get_gpg_sign(ignore_errors: bool = False):
    return exec_cmd("git config commit.gpgSign", ignore_errors=ignore_errors).strip()


def get_default_branch(ignore_errors: bool = False):
    return exec_cmd("git config init.defaultBranch", ignore_errors=ignore_errors).strip()
